/*
 * Sample Program to communicate with NXT via Bluetooth
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// Socket, used for Bluetooth rfcomm 
#include <sys/socket.h>
#include <sys/types.h>

// Bluetooth headers
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>

#define MAX_MESSAGE_SIZE	64

// Global Variables
int nxtSocket;

/****************************************************************
 * Initiate Bluetooth
 *	This will open a Bluetooth connection to the NXT
 * Returns: Failed (< 0)
 ***************************************************************/
int init_bluetooth(char *btAddress)	{
	struct sockaddr_rc addr={0};
	int status;

	/*-----------------------------------------------------------
	 * SOCK_STREAM
	 *	 Provides sequenced, reliable, two-way, connection-based 
	 *	  byte streams. An out-of-band data transmission 
	 *	  mechanism may be supported.
	 *----------------------------------------------------------*/

	// Allocate a socket
	nxtSocket = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);

	// Set what type and who to connect to
	addr.rc_family = AF_BLUETOOTH;
	addr.rc_channel = (uint8_t) 1;
	str2ba(btAddress, &addr.rc_bdaddr);

	// Connect to the NXT
	status = connect(nxtSocket, (struct sockaddr *)&addr, sizeof(addr) );

	if (status < 0)	{
		perror("Error connecting Bluetooth");
		return status;
	}

	return 0;
}

/****************************************************************
 * NXT Get Battery Level
 *	This will get the battery level on the NXT
 * Returns: The battery level as an integer 
 ***************************************************************/
int nxt_getBattery(void)	{
	/*------------------------------------------------------
	 * Direct Command format:
	 *  {Length/LSB, Length/MSB, Byte 0, Byte 1... Byte N}
	 *
	 * For GETBATTERYLEVEL (see direct commands):
	 *	Byte 0: 0x00
	 *	Byte 1: 0x0B
	 *	Length/LSB: 0x02, the command length is 2 bytes
	 *----------------------------------------------------*/
	char cmd[4]={0x02, 0x00, 0x00, 0x0B};
	char reply[MAX_MESSAGE_SIZE];
	int result;
	int bLevel;
	int replyLength;
	int error = 0;

	//- Send Request -----------------------------------------
	
	if ( (result = write(nxtSocket, cmd, 4)) < 0 )	{
		perror("Error sending GETBATTERYLEVEL command ");
		return result;
	}

	//- Read Reply -----------------------------------------
	// Get Bluetooth message length
	if ( (result = read(nxtSocket, reply, 2)) < 0)	{
		perror("Error receiving GETBATTERYLEVEL reply ");
		return result;
	}
	replyLength = reply[0] + (reply[1] * 256);
	
	// Get Return Package
	if ( (result = read(nxtSocket, reply, replyLength)) < 0)	{
		perror("Error receiving GETBATTERYLEVEL reply ");
		return result;
	}

	// Quick check to make sure we got everything
	if (replyLength != result)	{
		fprintf(stderr, 
				"GETBATTERYLEVEL : lengths do not match : %d != %d\n", 
				replyLength, result);
	}

	/*------------------------------------------------------
	 * Return Package format:
	 *	{Length/LSB, Length/MSB, Byte0, Byte1..., ByteN}
	 * For GETBATTERYLEVEL:
	 *	Byte0: 0x02
	 *	Byte1: 0x0B
	 *	Byte2: Status Byte
	 *	Byte3-4: Voltage in millivolts (UWORD)
	 *	Length/LSB: 0x05
	 *----------------------------------------------------*/
	// Byte 0
	if (reply[0] != 0x02)	{
		fprintf(stderr, "GETBATTERYLEVEL : Byte 0 : %hhX != 0x02\n", reply[0]);
		error = 1;
	}
	// Byte 1
	if (reply[1] != 0x13)	{
		fprintf(stderr, "GETBATTERYLEVEL : Byte 1 : %hhX != 0x13\n", reply[1]);
		error = 1;
	}
	// Byte 2
	if (reply[2] != 0x00)	{
		fprintf(stderr, "GETBATTERYLEVEL : Byte 2, Status : %hhX \n", reply[2]);
		error = 1;
	}
	if (error)	{
		return -1;
	}

	// Byte 3-4
	bLevel = reply[5] + (reply[6] * 256);
	return bLevel;
}


int main (void)	{
	// NXT Brick ALPHA Bluetooth address
	char btAddress[18] = "00:16:53:01:2C:84"; 
	int bLevel;

	// Initiate Bluetooth connection	
	if (init_bluetooth(btAddress) < 0)	{
		close(nxtSocket);
		return 1;
	}
	printf("Bluetooth connected to %s \n", btAddress); 

	// Get Battery level (direct command)
	bLevel = nxt_getBattery();
	if (bLevel < 0)	{
		close(nxtSocket);
		return 1;
	}
	printf("Battery Level: %.2f\n", bLevel/100.00);

	close(nxtSocket);
	return 0;
}
