/*
 * Sample Program to communicate with NXT via Bluetooth
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// Socket, used for Bluetooth rfcomm 
#include <sys/socket.h>
#include <sys/types.h>

// Bluetooth headers
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>

#define MAX_MESSAGE_SIZE	64
#define NXT_MAILBOX_EMPTY	0x40

// Global Variables
int nxtSocket;

/****************************************************************
 * Initiate Bluetooth
 *	This will open a Bluetooth connection to the NXT
 * Returns: Failed (< 0)
 ***************************************************************/
int init_bluetooth(char *btAddress)	{
	struct sockaddr_rc addr={0};
	int status;

	/*-----------------------------------------------------------
	 * SOCK_STREAM
	 *	 Provides sequenced, reliable, two-way, connection-based 
	 *	  byte streams. An out-of-band data transmission 
	 *	  mechanism may be supported.
	 *----------------------------------------------------------*/

	// Allocate a socket
	nxtSocket = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);

	// Set what type and who to connect to
	addr.rc_family = AF_BLUETOOTH;
	addr.rc_channel = (uint8_t) 1;
	str2ba(btAddress, &addr.rc_bdaddr);

	// Connect to the NXT
	status = connect(nxtSocket, (struct sockaddr *)&addr, sizeof(addr) );

	if (status < 0)	{
		perror("Error connecting Bluetooth");
		return status;
	}

	return 0;
}

/****************************************************************
 * NXT Read Message
 *	This will read a mailbox message on the NXT. Usually the
 *	 message is a reply from a sent message.
 * Parameters
 *	mbox: mailbox to read from
 *	message: return variable containing the message
 ***************************************************************/
int nxt_readMessage(int mbox, char **message)	{
	unsigned char btLength[2]={0x00,0x00};
	unsigned char cmd[5]={0x0};
	unsigned char reply[MAX_MESSAGE_SIZE];
	int result;
	int cmdLength;
	int msgSize;
	int error = 0;


	//- Send Request -----------------------------------------

	/*------------------------------------------------------
	 * Direct Command format:
	 *  {Length/LSB, Length/MSB, Byte0, Byte1 ... , ByteN}
	 *
	 * For MESSAGEREAD (see direct commands):
	 *	Byte 0: 0x00 or 0x80
	 *	Byte 1: 0x13
	 *	Byte 2: Remote Inbox number (0-9)
	 *	Byte 3: Local Inbox number (0-9)
	 *	Byte 4: Remove? 
	 *	Length/LSB: 5
	 *----------------------------------------------------*/

	// NXT Command
	cmd[0] = 0x00;
	cmd[1] = 0x13;
	cmd[2] = mbox+10;
	cmd[3] = 0x00;
	cmd[4] = 0x01;

	// Bluetooth message length
	btLength[0] = 5;

	// Send Bluetooth length
	if ( (result = write(nxtSocket, btLength, 2)) < 0)	{
		perror("Error sending MESSAGEREAD command ");
		return result;
	}
	
	// Send Command
	if ( (result = write(nxtSocket, cmd, 5)) < 0)	{
		perror("Error sending MESSAGEREAD command ");
		return result;
	}


	//- Read Reply -----------------------------------------
	// Get Bluetooth message length
	if ( (result = read(nxtSocket, reply, 2)) < 0)	{
		perror("Error receiving MESSAGEREAD reply ");
		return result;
	}
	cmdLength = reply[0] + (reply[1] * 256);
	
	// Get Return Package
	if ( (result = read(nxtSocket, reply, cmdLength)) < 0)	{
		perror("Error recieveing MESSAGEREAD reply ");
		return result;
	}

	// Quick check to make sure we got everything
	if (cmdLength != result)	{
		fprintf(stderr, 
				"MESSAGEREAD : lengths do not match : %d != %d\n", 
				cmdLength, result);
	}

	/*------------------------------------------------------
	 * Return Package for MESSAGEREAD
	 * Byte 0: 0x02
	 * Byte 1: 0x13
	 * Byte 2: Status Byte
	 * Byte 3: Local inbox number (0-9), should match request
	 * Byte 4: Message Size
	 * Byte 5-63: Message data (padded)
	 *-----------------------------------------------------*/
	// Byte 0
	if (reply[0] != 0x02)	{
		fprintf(stderr, "MESSAGEREAD : Byte 0 : %hhX != 0x02\n", reply[0]);
		error = 1;
	}
	// Byte 1
	if (reply[1] != 0x13)	{
		fprintf(stderr, "MESSAGEREAD : Byte 1 : %hhX != 0x13\n", reply[1]);
		error = 1;
	}
	// Byte 2
	if (reply[2] == NXT_MAILBOX_EMPTY)	{
		printf("Mailbox Empty");
		message = NULL;
		return reply[2];
	}
	if (reply[2] != 0x00)	{
		fprintf(stderr, "MESSAGEREAD : Byte 2, Status : %hhX \n", reply[2]);
		error = 1;
	}
	// Byte 3
	if (reply[3] != 0x00)	{
		fprintf(stderr, "MESSAGEREAD : Byte 3, mbox: %hhX != 0x00\n", reply[3]);
		error = 1;
	}
	
	if (error)	{
		message = NULL;
		return -1;
	}

	// Byte 4
	msgSize = reply[4];
	*message = (char *)malloc(sizeof(char) * msgSize + 2);
	printf("reply[5-%d]: %s\n", msgSize+6, &reply[5]);

	// Byte 5-63: Message data
	memcpy(*message, &reply[5], msgSize);	
	printf("message: %s\n", *message);

	return 0;
}


/****************************************************************
 * NXT Send Message
 *	This will send a message to a specific mailbox on the NXT
 * Parameters
 *	mbox: mailbox to write to on the NXT
 *	message: message to send to NXT
 ***************************************************************/
int nxt_sendMessage(int mbox, char *message)	{
	unsigned char btLength[2]={0x00,0x00};
	unsigned char cmd[MAX_MESSAGE_SIZE]={0x0};
	unsigned char reply[MAX_MESSAGE_SIZE]={0x0};
	int result;
	int msgSize;
	int replyLength;
	int error = 0;

	//- Send Request -----------------------------------------

	/*------------------------------------------------------
	 * Direct Command format:
	 *  {Length/LSB, Length/MSB, Byte 0, Byte 1 ... , Byte N}
	 *
	 * For MESSAGEWRITE (see direct commands):
	 *	Byte 0: 0x00 or 0x80
	 *	Byte 1: 0x09
	 *	Byte 2: Inbox number (0-9)
	 *	Byte 3: Message Size
	 *	Byte 4-N: Message Data, where N = Message size + 3
	 *	Length/LSB: Message Size + 4
	 *
	 *	Max Message Size: 59 (Max packet length is 64 bytes) 
	 *----------------------------------------------------*/

	//Create the MESSAGEWRITE command
	msgSize = strlen(message) + 1; // Add one for null termination
	if (msgSize > (MAX_MESSAGE_SIZE - 4) )	{
		fprintf(stderr, "MESSAGEWRITE : Message is to long");
		return -1;
	}

	// NXT Command
	cmd[0] = 0x00;
	cmd[1] = 0x09;
	cmd[2] = mbox;
	cmd[3] = msgSize;
	memcpy(&cmd[4], message, msgSize);

	// Bluetooth length
	btLength[0]= 4 + msgSize;
	
	// Send Bluetooth length
	if ( (result = write(nxtSocket, btLength, 2)) < 0)	{
		perror("Error sending MESSAGEWRITE command ");
		return result;
	}

	// Send Command 
	if ( (result = write(nxtSocket, cmd, btLength[0])) < 0)	{
		perror("Error sending MESSAGEWRITE command ");
		return result;
	}

	//- Read Reply -----------------------------------------
	
	// Get Bluetooth message length
	if ( (result = read(nxtSocket, reply, 2)) < 0)	{
		perror("Error receiving MESSAGEWRITE reply ");
		return result;
	}
	replyLength = reply[0] + (reply[1] * 256);
	
	// Get Return Package
	if ( (result = read(nxtSocket, reply, replyLength)) < 0)	{
		perror("Error receiving MESSAGEWRITE reply ");
		return result;
	}

	// Quick check to make sure we got everything
	if (replyLength != result)	{
		fprintf(stderr, 
				"MESSAGEWRITE : lengths do not match : %d != %d\n", 
				replyLength, result);
	}

	/*------------------------------------------------------
	 * Return Package for MESSAGEWRITE
	 * Byte 0: 0x02
	 * Byte 1: 0x09
	 * Byte 2: Status Byte
	 * Byte 3: Local inbox number (0-9), should match request
	 * Byte 4: Message Size
	 * Byte 5-63: Message data (padded)
	 *----------------------------------------------------*/
	// Byte 0
	if (reply[0] != 0x02)	{
		fprintf(stderr, "MESSAGEWRITE : Byte 0 : %hhX != 0x02\n", reply[0]);
		error = 1;
	}
	// Byte 1
	if (reply[1] != 0x09)	{
		fprintf(stderr, "MESSAGEWRITE : Byte 1 : %hhX != 0x13\n", reply[1]);
		error = 1;
	}
	// Byte 2
	if (reply[2] != 0x00)	{
		fprintf(stderr, "MESSAGEWRITE : Byte 2, Status : %hhX \n", reply[2]);
		error = 1;
	}

	if (error)	{
		return -1;
	}

	return 0;
}

int main (void)	{
	// NXT Brick ALPHA Bluetooth address
	char btAddress[18] = "00:16:53:01:2C:84"; 
	int status;
	char *reply;

	// Initiate Bluetooth connection	
	if (init_bluetooth(btAddress) < 0)	{
		close(nxtSocket);
		return 1;
	}
	printf("Bluetooth connected to %s \n", btAddress); 

	// Send the Message
	nxt_sendMessage(3, "world");

	// Read the NXT reply message
	while ( (status = nxt_readMessage(3, &reply)) == NXT_MAILBOX_EMPTY); 
	if (status < 0)	{
		return 1;
	}
	printf("Reply: %s\n", reply);
	free(reply);

	close(nxtSocket);
	return 0;
}
